import 'package:flutter/material.dart';
import 'package:auto_size_text/auto_size_text.dart';
import 'package:marquee/marquee.dart' as marquee;


class TileDimensions {
  double width;
  double height;
  double padding;
  int elementsPerRow;
  TileDimensions({required this.width, required this.height, required this.elementsPerRow, required this.padding});
}


TileDimensions getTileDimensions(BuildContext context, BoxConstraints preferredConstraints, {double padding: 8.0, double preferredWidth: 156.0, double preferredHeight: 242.0}) {
  int elementsPerRow = (preferredConstraints.maxWidth - 2 * padding) ~/ (preferredWidth + padding);
  double width = (preferredConstraints.maxWidth - 2 * padding - (elementsPerRow - 1) * padding) / elementsPerRow;
  double height = width * preferredHeight / preferredWidth;
  return new TileDimensions(width: width, height: height, elementsPerRow: elementsPerRow, padding: padding);
}


List<Widget> tileWidgets({
  required BuildContext context,
  required TileDimensions tileDimensions,
  required int widgetCount,
  required Widget Function(BuildContext context, int index) builder,
  Widget? leadingSubHeader,
  Widget? subHeader,
  Widget? leadingWidget
}) {
  List<Widget> widgets = <Widget>[];
  if (leadingSubHeader != null) widgets.add(leadingSubHeader);
  if (leadingWidget != null) widgets.add(leadingWidget);
  if (subHeader != null) widgets.add(subHeader);
  int rowIndex = 0;
  List<Widget> rowChildren = <Widget>[];
  for (int index = 0; index < widgetCount; index++) {
    rowChildren.add(
      builder(context, index),
    );
    rowIndex++;
    if (rowIndex > tileDimensions.elementsPerRow - 1) {
      widgets.add(
        new Container(
          height: tileDimensions.height + tileDimensions.padding,
          margin: EdgeInsets.only(left: tileDimensions.padding, right: tileDimensions.padding),
          alignment: Alignment.topCenter,
          child: Row(
            mainAxisSize: MainAxisSize.max,
            mainAxisAlignment: MainAxisAlignment.spaceBetween,
            crossAxisAlignment: CrossAxisAlignment.start,
            children: rowChildren,
          ),
        ),
      );
      rowIndex = 0;
      rowChildren = <Widget>[];
    }
  }
  if (widgetCount % tileDimensions.elementsPerRow != 0) {
    rowChildren = <Widget>[];
    for (int index = widgetCount - (widgetCount % tileDimensions.elementsPerRow); index < widgetCount; index++) {
      rowChildren.add(
        builder(context, index),
      );
    }
    for (int index = 0; index < tileDimensions.elementsPerRow - (widgetCount % tileDimensions.elementsPerRow); index++) {
      rowChildren.add(
        Container(
          height: tileDimensions.height,
          width: tileDimensions.width,
        ),
      );
    }
    widgets.add(
      new Container(
        height: tileDimensions.height + tileDimensions.padding,
        margin: EdgeInsets.only(left: tileDimensions.padding, right: tileDimensions.padding),
        alignment: Alignment.topCenter,
        child: Row(
          mainAxisSize: MainAxisSize.max,
          mainAxisAlignment: MainAxisAlignment.spaceBetween,
          crossAxisAlignment: CrossAxisAlignment.start,
          children: rowChildren,
        ),
      ),
    );
  }
  return widgets;
}


class SubHeader extends StatelessWidget {
  final String? text;
  final TextStyle? style;
  const SubHeader(this.text, {Key? key, this.style}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Container(
      alignment: Alignment.centerLeft,
      height: 48,
      padding: EdgeInsets.fromLTRB(16, 0, 0, 0),
      color: Theme.of(context).scaffoldBackgroundColor,
      child: Text(
        text!,
        style: this.style ?? Theme.of(context).textTheme.headline5,
      ),
    );
  }
}


class FadeFutureBuilder extends StatefulWidget {
  final Future<Object> Function() future;
  final Widget Function(BuildContext context) initialWidgetBuilder;
  final Widget Function(BuildContext context, Object? object) finalWidgetBuilder;
  final Widget Function(BuildContext context, Object object) errorWidgetBuilder;
  final Duration transitionDuration;

  const FadeFutureBuilder({
    Key? key,
    required this.future,
    required this.initialWidgetBuilder,
    required this.finalWidgetBuilder,
    required this.errorWidgetBuilder,
    required this.transitionDuration,
  }) : super(key: key);
  FadeFutureBuilderState createState() => FadeFutureBuilderState();
}


class FadeFutureBuilderState extends State<FadeFutureBuilder>
    with SingleTickerProviderStateMixin {
  bool _init = true;
  Widget _currentWidget = Container();
  late AnimationController _widgetOpacityController;
  late Animation<double> _widgetOpacity;
  Object? _futureResolve;

  @override
  void initState() {
    super.initState();
    this._currentWidget = widget.initialWidgetBuilder(context);
  }

  @override
  void didChangeDependencies() async {
    super.didChangeDependencies();
    if (this._init) {
      this._currentWidget = widget.initialWidgetBuilder(context);
      this._widgetOpacityController = new AnimationController(
        vsync: this,
        duration: widget.transitionDuration,
        reverseDuration: widget.transitionDuration,
      );
      this._widgetOpacity = new Tween<double>(
        begin: 1.0,
        end: 0.0,
      ).animate(new CurvedAnimation(
        parent: this._widgetOpacityController,
        curve: Curves.easeInOutCubic,
        reverseCurve: Curves.easeInOutCubic,
      ));
      try {
        this._futureResolve = await widget.future();
        this._widgetOpacityController.forward();
        Future.delayed(widget.transitionDuration, () {
          this.setState(() {
            this._currentWidget =
                widget.finalWidgetBuilder(context, this._futureResolve);
          });
          this._widgetOpacityController.reverse();
        });
      } catch (exception) {
        this._widgetOpacityController.forward();
        Future.delayed(widget.transitionDuration, () {
          this.setState(() {
            this._currentWidget = widget.errorWidgetBuilder(context, exception);
          });
          this._widgetOpacityController.reverse();
        });
      }
      this._init = false;
    }
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      child: FadeTransition(
        opacity: this._widgetOpacity,
        child: this._currentWidget,
      ),
    );
  }
}


class ExceptionWidget extends StatelessWidget {
  final EdgeInsets margin;
  final double height;
  final Icon? icon;
  final String? title;
  final TextStyle? titleStyle;
  final String? subtitle;
  final TextStyle? subtitleStyle;
  final String? assetImage;
  final bool large;
  final Color? cardColor;

  const ExceptionWidget({
    Key? key,
    this.assetImage,
    this.icon,
    required this.margin,
    required this.height,
    required this.title,
    required this.titleStyle,
    required this.subtitle,
    required this.subtitleStyle,
    this.cardColor,
    this.large: false
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return this.large ? Card(
      color: this.cardColor ?? Theme.of(context).cardColor,
      elevation: 2.0,
      clipBehavior: Clip.antiAlias,
      margin: this.margin,
      child: Container(
        width: MediaQuery.of(context).size.width - 16.0,
        child: Column(
          mainAxisAlignment: MainAxisAlignment.start,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: [
            this.assetImage != null ? Image.asset(
              this.assetImage!,
              height: this.height,
              width: MediaQuery.of(context).size.width - 16.0,
              fit: BoxFit.fitWidth,
              alignment: Alignment.center,
            ): Container(),
            Container(
              margin: EdgeInsets.symmetric(
                horizontal: 16.0,
                vertical: 4.0,
              ),
              width: MediaQuery.of(context).size.width - 16.0,
              child: Column(
                crossAxisAlignment: CrossAxisAlignment.start,
                mainAxisAlignment: MainAxisAlignment.center,
                children: [
                  Text(
                    this.title!,
                    style: this.titleStyle ?? Theme.of(context).textTheme.headline1,
                    textAlign: TextAlign.start,
                  ),
                  Divider(
                    color: Colors.transparent,
                    height: 4.0,
                  ),
                  Text(
                    this.subtitle!,
                    style: this.subtitleStyle ?? Theme.of(context).textTheme.headline5,
                    textAlign: TextAlign.start,
                  ),
                ],
              ),
            ),
          ],
        ),
      ),
    ): Card(
      color: this.cardColor ?? Theme.of(context).cardColor,
      elevation: 2.0,
      clipBehavior: Clip.antiAlias,
      margin: this.margin,
      child: Container(
        width: MediaQuery.of(context).size.width - 16,
        child: Row(
          mainAxisAlignment: MainAxisAlignment.start,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: [
            this.assetImage != null ? Image.asset(
              this.assetImage!,
              height: this.height,
              width: this.height,
              fit: BoxFit.cover,
              alignment: Alignment.center,
            ): Container(),
            Container(
              margin: EdgeInsets.only(left: 8, right: 8),
              width: MediaQuery.of(context).size.width - 32 - this.height,
              child: Column(
                crossAxisAlignment: CrossAxisAlignment.start,
                mainAxisAlignment: MainAxisAlignment.center,
                children: [
                  Text(
                    this.title!,
                    style: this.titleStyle ?? Theme.of(context).textTheme.headline1,
                    textAlign: TextAlign.start,
                  ),
                  Divider(
                    color: Colors.transparent,
                    height: 4.0,
                  ),
                  Text(
                    this.subtitle!,
                    style: this.subtitleStyle ?? Theme.of(context).textTheme.headline5,
                    textAlign: TextAlign.start,
                  ),
                ],
              ),
            ),
          ],
        ),
      ),
    );
  }
}


class FakeLinearProgressIndicator extends StatelessWidget {
  final String label;
  final TextStyle? style;
  final Duration duration;
  final double? width;
  final EdgeInsets? margin;
  final Animation<Color>? valueColor;
  final Color? backgroundColor;

  FakeLinearProgressIndicator({
    Key? key,
    required this.label,
    required this.duration,
    this.width,
    this.margin,
    this.style,
    this.valueColor,
    this.backgroundColor
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return TweenAnimationBuilder(
      tween: Tween<double>(begin: 0.0, end: 1.0),
      duration: this.duration,
      child: Text(this.label),
      curve: Curves.linear,
      builder: (BuildContext context, double value, Widget? child) => Center(
        child: Container(
          margin: this.margin ?? EdgeInsets.zero,
          alignment: Alignment.center,
          width: this.width ?? 148.0,
          child: Column(
            mainAxisSize: MainAxisSize.min,
            mainAxisAlignment: MainAxisAlignment.start,
            crossAxisAlignment: CrossAxisAlignment.center,
            children: [
              Text(
                this.label,
                style: this.style ?? Theme.of(context).textTheme.headline4,
              ),
              Divider(
                height: 12.0,
                color: Colors.transparent,
              ),
              LinearProgressIndicator(value: value, valueColor: this.valueColor, backgroundColor: this.backgroundColor),
            ],
          ),
        ),
      ),
    );
  }
}


class ClosedTile extends StatelessWidget {
  final String? title;
  final String? subtitle;
  final Color? cardColor;
  const ClosedTile({Key? key, required this.open, required this.title, required this.subtitle, this.cardColor}) : super(key: key);

  final Function open;

  @override
  Widget build(BuildContext context) {
    return Card(
      margin: EdgeInsets.only(
        left: 8.0,
        right: 8.0,
        top: 4.0,
        bottom: 4.0,
      ),
      color: this.cardColor ?? Theme.of(context).cardColor,
      elevation: 2.0,
      child: ListTile(
        title: Text(this.title!),
        subtitle: Text(this.subtitle!),
        onTap: open as void Function()?,
      ),
    );
  }
}


class Marquee extends StatelessWidget {
  const Marquee({
    Key? key,
    required this.text,
    required this.style,
    this.velocity = 20.0,
    this.blankSpace = 40.0,
    this.startAfter = const Duration(seconds: 2),
    this.pauseAfterRound = const Duration(seconds: 2),
  }) : super(key: key);

  final String text;
  final TextStyle style;
  final double velocity;
  final double blankSpace;
  final Duration startAfter;
  final Duration pauseAfterRound;

  @override
  Widget build(BuildContext context) {
    return Container(
      alignment: Alignment.centerLeft,
      height: (style.fontSize! + 4.0) * MediaQuery.of(context).textScaleFactor,
      child: AutoSizeText(
        text,
        minFontSize: style.fontSize!,
        maxFontSize: style.fontSize!,
        style: style,
        overflowReplacement: marquee.Marquee(
          text: text,
          blankSpace: blankSpace,
          accelerationCurve: Curves.easeOutCubic,
          velocity: velocity,
          startPadding: 2.0,
          startAfter: startAfter,
          pauseAfterRound: pauseAfterRound,
          style: style,
        ),
      ),
    );
  }
}
